// qtractorMidiControl.h
//
/****************************************************************************
   Copyright (C) 2005-2009, rncbc aka Rui Nuno Capela. All rights reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*****************************************************************************/

#ifndef __qtractorMidiControl_h
#define __qtractorMidiControl_h

#include "qtractorObserver.h"

#include <QEvent>


//----------------------------------------------------------------------
// qtractorMidiControlEvent - MIDI Control custom event.
//

class qtractorMidiControlEvent : public QEvent
{
public:

	// Contructor.
	qtractorMidiControlEvent(QEvent::Type eType, unsigned short iChannel,
		unsigned char controller, unsigned char value)
		: QEvent(eType), m_channel(iChannel),
			m_controller(controller), m_value(value) {}

	// Accessors.
	unsigned short channel()    const { return m_channel; }
	unsigned short controller() const { return m_controller; }
	unsigned short value()      const { return m_value; }

private:

	// Instance variables.
	unsigned short m_channel;
	unsigned short m_controller;
	unsigned short m_value;
};


//----------------------------------------------------------------------
// class qtractorMidiControlWidget -- Template widget observer/visitor.
//

#include <QCheckBox>
#include <QSpinBox>
#include <QSlider>


template <class W>
class qtractorMidiControlWidget : public W
{
public:

	// Local observer.
	class Observer : public qtractorObserver
	{
	public:
		// Constructor.
		Observer(qtractorSubject *pSubject, W *pWidget)
			: qtractorObserver(pSubject), m_pWidget(pWidget) {}

	protected:

		// Visitors overload.
		void visit(QCheckBox *pCheckBox, float fValue)
			{ pCheckBox->setChecked(fValue > 0.0f); }
		void visit(QSpinBox *pSpinBox, float fValue)
			{ pSpinBox->setValue(int(100.0f * fValue)); }
		void visit(QDoubleSpinBox *pDoubleSpinBox, float fValue)
			{ pDoubleSpinBox->setValue(fValue); }
		void visit(QSlider *pSlider, float fValue)
			{ pSlider->setValue(int(10000.0f * fValue)); }

		// Observer updater.
		void update() { visit(m_pWidget, value()); }

	private:
		// Members.
		W *m_pWidget;
	};

	// Constructor.
	qtractorMidiControlWidget(QWidget *pParent = 0)
		: W(pParent), m_pObserver(NULL) {};

	// Destructor.
	~qtractorMidiControlWidget()
		{ if (m_pObserver) delete m_pObserver; }

	// Setup.
	void setSubject(qtractorSubject *pSubject)
	{
		if (m_pObserver) delete m_pObserver;
		m_pObserver = new Observer(pSubject, this); 
	}

	// Observer accessor.
	Observer *observer() const { return m_pObserver; }

private:

	// Members.
	Observer *m_pObserver;
};


//----------------------------------------------------------------------
// class qtractorMidiControlCheckBox -- Concrete widget observer.
//

class qtractorMidiControlCheckBox : public qtractorMidiControlWidget<QCheckBox>
{
public:

	// Constructor.
	qtractorMidiControlCheckBox(QWidget *pParent = 0) 
		: qtractorMidiControlWidget<QCheckBox> (pParent) {}
};


//----------------------------------------------------------------------
// class qtractorMidiControlSpinBox -- Concrete widget observer.
//

class qtractorMidiControlSpinBox : public qtractorMidiControlWidget<QSpinBox>
{
public:

	// Constructor.
	qtractorMidiControlSpinBox(QWidget *pParent = 0) 
		: qtractorMidiControlWidget<QSpinBox> (pParent) {}
};


//----------------------------------------------------------------------
// class qtractorMidiControlDoubleSpinBox -- Concrete widget observer.
//

class qtractorMidiControlDoubleSpinBox : public qtractorMidiControlWidget<QDoubleSpinBox>
{
public:

	// Constructor.
	qtractorMidiControlDoubleSpinBox(QWidget *pParent = 0) 
		: qtractorMidiControlWidget<QDoubleSpinBox> (pParent) {}
};


//----------------------------------------------------------------------
// class qtractorMidiControlSlider -- Concrete widget observer.
//

class qtractorMidiControlSlider : public qtractorMidiControlWidget<QSlider>
{
public:

	// Constructor.
	qtractorMidiControlSlider(QWidget *pParent = 0) 
		: qtractorMidiControlWidget<QSlider> (pParent) {}
};


//----------------------------------------------------------------------
// class qtractorMidiControlObserver -- MIDI controller observers.
//

class qtractorMidiControlObserver : public qtractorObserver
{
public:

	// Constructor.
	qtractorMidiControlObserver(qtractorSubject *pSubject,
		unsigned short iChannel, unsigned short iController);

	// Destructor.
	~qtractorMidiControlObserver();

	// Key accessors.
	unsigned short channel()    const { return m_iChannel; }
	unsigned short controller() const { return m_iController; }

	// MIDI controller dispatcher.
	static void dispatch(unsigned short iChannel,
		unsigned short iController, unsigned short iValue);

protected:

	// Updater.
	void update();

private:

	// Members.
	unsigned short m_iChannel;
	unsigned short m_iController;

	static QList<qtractorMidiControlObserver *> g_midiControlObservers;
};


#endif  // __qtractorMidiControl_h


// end of qtractorMidiControl.h
