// qtractorMidiControlTestForm.cpp
//
/****************************************************************************
   Copyright (C) 2005-2009, rncbc aka Rui Nuno Capela. All rights reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*****************************************************************************/

#include "qtractorMidiControlTestForm.h"

#include "qtractorMidiControl.h"
#include "qtractorObserver.h"

#include <QThread>
#include <QTimer>

// Timer constant (magic) stuff.
#define QTRACTOR_TIMER_MSECS    66

// Specialties for thread-callback comunication.
#define QTRACTOR_CTL_EVENT      QEvent::Type(QEvent::User + 1)


//----------------------------------------------------------------------
// class qtractorMidiControlInputThread -- MIDI input thread (singleton).
//

class qtractorMidiControlInputThread : public QThread
{
public:

	// Constructor.
	qtractorMidiControlInputThread(qtractorMidiControlTestForm *pTestForm)
		: QThread(), m_pTestForm(pTestForm), m_bRunState(false) {}

	// Run-state accessors.
	void setRunState(bool bRunState)
		{ m_bRunState = bRunState; }
	bool runState() const
		{ return m_bRunState; }

protected:

	// The main thread executive.
	void run()
	{
		snd_seq_t *pAlsaSeq = m_pTestForm->alsaSeq();
		if (pAlsaSeq == NULL)
			return;

		int nfds;
		struct pollfd *pfds;

		nfds = snd_seq_poll_descriptors_count(pAlsaSeq, POLLIN);
		pfds = (struct pollfd *) alloca(nfds * sizeof(struct pollfd));
		snd_seq_poll_descriptors(pAlsaSeq, pfds, nfds, POLLIN);

		m_bRunState = true;

		int iPoll = 0;
		while (m_bRunState && iPoll >= 0) {
			// Wait for events...
			iPoll = poll(pfds, nfds, 200);
			while (iPoll > 0) {
				snd_seq_event_t *pEv = NULL;
				snd_seq_event_input(pAlsaSeq, &pEv);
				// Process input event - ...
				// - enqueue to input track mapping;
				m_pTestForm->capture(pEv);
			//	snd_seq_free_event(pEv);
				iPoll = snd_seq_event_input_pending(pAlsaSeq, 0);
			}
		}
	}

private:

	// The thread launcher engine.
	qtractorMidiControlTestForm *m_pTestForm;

	// Whether the thread is logically running.
	bool m_bRunState;
};


//----------------------------------------------------------------------------
// qtractorMidiControlTestForm -- UI wrapper form.

// Pseudo-singleton reference.
qtractorMidiControlTestForm *qtractorMidiControlTestForm::g_pTestForm = NULL;

// Constructor.
qtractorMidiControlTestForm::qtractorMidiControlTestForm (
	QWidget *pParent, Qt::WindowFlags wflags )
	: QDialog(pParent, wflags)
{
	// Setup UI struct...
	m_ui.setupUi(this);

	// Set pseudo-singleton reference.
	g_pTestForm = this;

	m_pAlsaSeq    = NULL;
	m_iAlsaClient = -1;
	m_iAlsaPort   = -1;

	m_pInputThread = NULL;

	// Initialize subjects.
	m_pBooleanSubject = new qtractorSubject("Boolean");
	m_pIntegerSubject = new qtractorSubject("Integer");
	m_pLinearSubject  = new qtractorSubject("Linear");
	m_pLogSubject     = new qtractorSubject("Log");

	// Observers.
	m_ui.BooleanCheckBox->setSubject(m_pBooleanSubject);
	m_ui.IntegerSpinBox->setSubject(m_pIntegerSubject);
	m_ui.LinearSpinBox->setSubject(m_pLinearSubject);
	m_ui.LinearSlider->setSubject(m_pLinearSubject);
	m_ui.LogSpinBox->setSubject(m_pLogSubject);
	m_ui.LogSlider->setSubject(m_pLogSubject);
	m_ui.LogSlider->observer()->setLogarithmic(true);

	// MIDI Observers.
	m_pBooleanMidiObserver
		= new qtractorMidiControlObserver(m_pBooleanSubject, 0, 1);
	m_pIntegerMidiObserver
		= new qtractorMidiControlObserver(m_pIntegerSubject, 0, 10);
	m_pLinearMidiObserver
		= new qtractorMidiControlObserver(m_pLinearSubject, 0, 16);
	m_pLogMidiObserver
		= new qtractorMidiControlObserver(m_pLogSubject, 0, 7);
	m_pLogMidiObserver->setLogarithmic(true);

	// Initialize local widgets.
	m_ui.IntegerSpinBox->setMinimum(0);
	m_ui.IntegerSpinBox->setMaximum(100);
	m_ui.IntegerSpinBox->setSingleStep(1);
	m_ui.IntegerSpinBox->setAccelerated(true);

	m_ui.LinearSpinBox->setMinimum(0.0f);
	m_ui.LinearSpinBox->setMaximum(1.0f);
	m_ui.LinearSpinBox->setSingleStep(0.01f);
	m_ui.LinearSpinBox->setAccelerated(true);

	m_ui.LinearSlider->setMinimum(0);
	m_ui.LinearSlider->setMaximum(10000);
	m_ui.LinearSlider->setPageStep(1000);
	m_ui.LinearSlider->setSingleStep(100);

	m_ui.LogSpinBox->setMinimum(0.0f);
	m_ui.LogSpinBox->setMaximum(1.0f);
	m_ui.LogSpinBox->setSingleStep(0.01f);
	m_ui.LogSpinBox->setAccelerated(true);

	m_ui.LogSlider->setMinimum(0);
	m_ui.LogSlider->setMaximum(10000);
	m_ui.LogSlider->setPageStep(1000);
	m_ui.LogSlider->setSingleStep(100);

	// Open new ALSA sequencer client...
	if (snd_seq_open(&m_pAlsaSeq, "hw", SND_SEQ_OPEN_DUPLEX, 0) >= 0) {
		// Set client identification...
		snd_seq_set_client_name(m_pAlsaSeq, "qtractorMidiControlTest");
		m_iAlsaClient = snd_seq_client_id(m_pAlsaSeq);
		// Create duplex port
		m_iAlsaPort = snd_seq_create_simple_port(m_pAlsaSeq, "control",
			SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE |
			SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_SUBS_READ,
			SND_SEQ_PORT_TYPE_MIDI_GENERIC | SND_SEQ_PORT_TYPE_APPLICATION);
		// Create and start our own MIDI input queue thread...
		m_pInputThread = new qtractorMidiControlInputThread(this);
		m_pInputThread->start(QThread::TimeCriticalPriority);
	}

	QObject::connect(m_ui.BooleanCheckBox,
		SIGNAL(toggled(bool)),
		SLOT(booleanCheckBoxChanged(bool)));

	QObject::connect(m_ui.IntegerSpinBox,
		SIGNAL(valueChanged(const QString&)),
		SLOT(integerSpinBoxChanged(const QString&)));

	QObject::connect(m_ui.LinearSpinBox,
		SIGNAL(valueChanged(const QString&)),
		SLOT(linearSpinBoxChanged(const QString&)));
	QObject::connect(m_ui.LinearSlider,
		SIGNAL(valueChanged(int)),
		SLOT(linearSliderChanged(int)));

	QObject::connect(m_ui.LogSpinBox,
		SIGNAL(valueChanged(const QString&)),
		SLOT(logSpinBoxChanged(const QString&)));
	QObject::connect(m_ui.LogSlider,
		SIGNAL(valueChanged(int)),
		SLOT(logSliderChanged(int)));

	QObject::connect(m_ui.ClosePushButton,
		SIGNAL(clicked()),
		SLOT(close()));

	// Register the first timer slot.
	QTimer::singleShot(QTRACTOR_TIMER_MSECS, this, SLOT(timerSlot()));
}


qtractorMidiControlTestForm::~qtractorMidiControlTestForm (void)
{
	// Reset pseudo-singleton reference.
	g_pTestForm = NULL;

	// Last but not least, delete input thread...
	if (m_pInputThread) {
		// Try to terminate executive thread,
		// but give it a bit of time to cleanup...
		if (m_pInputThread->isRunning()) {
			m_pInputThread->setRunState(false);
		//	m_pInputThread->terminate();
			m_pInputThread->wait();
		}
		delete m_pInputThread;
		m_pInputThread = NULL;
	}

	if (m_pAlsaSeq) {
		snd_seq_delete_simple_port(m_pAlsaSeq, m_iAlsaPort);
		m_iAlsaPort   = -1;
		snd_seq_close(m_pAlsaSeq);
		m_iAlsaClient = -1;
		m_pAlsaSeq    = NULL;
	}

	// Destroy MIDI observers.
	delete m_pBooleanMidiObserver;
	delete m_pIntegerMidiObserver;
	delete m_pLinearMidiObserver;
	delete m_pLogMidiObserver;

	// Destroy subjects.
	delete m_pBooleanSubject;
	delete m_pIntegerSubject;
	delete m_pLinearSubject;
	delete m_pLogSubject;
}


// Pseudo-singleton reference (static).
qtractorMidiControlTestForm *qtractorMidiControlTestForm::getInstance (void)
{
	return g_pTestForm;
}


// ALSA sequencer client descriptor accessor.
snd_seq_t *qtractorMidiControlTestForm::alsaSeq (void) const
{
	return m_pAlsaSeq;
}

int qtractorMidiControlTestForm::alsaClient (void) const
{
	return m_iAlsaClient;
}

int qtractorMidiControlTestForm::alsaPort (void) const
{
	return m_iAlsaPort;
}


// MIDI event capture method.
void qtractorMidiControlTestForm::capture ( snd_seq_event_t *pEv )
{
	// Post controller event...
	if (pEv->type == SND_SEQ_EVENT_CONTROLLER) {
		QApplication::postEvent(this,
			new qtractorMidiControlEvent(
				QTRACTOR_CTL_EVENT,
				pEv->data.control.channel,
				pEv->data.control.param,
				pEv->data.control.value));
	}
}


// Custom event handler.
void qtractorMidiControlTestForm::customEvent ( QEvent *pEvent )
{
	if (pEvent->type() == QTRACTOR_CTL_EVENT) {
		// Contrller event handling...
		qtractorMidiControlEvent *pCtlEvent
			= static_cast<qtractorMidiControlEvent *> (pEvent);
		if (pCtlEvent) {
			receiveController(
				pCtlEvent->channel(),
				pCtlEvent->controller(),
				pCtlEvent->value());
		}
	}
}


// Send MIDI controller event.
void qtractorMidiControlTestForm::sendController ( unsigned short iChannel,
	unsigned short iController, unsigned short iValue ) const
{
	qDebug("sendController: channel %u, controller %u, value %u",
		iChannel, iController, iValue);

	// Initialize sequencer event...
	snd_seq_event_t ev;
	snd_seq_ev_clear(&ev);

	// Addressing...
	snd_seq_ev_set_source(&ev, m_iAlsaPort);
	snd_seq_ev_set_subs(&ev);

	// The event will be direct...
	snd_seq_ev_set_direct(&ev);

	// Set controller parameters...
	ev.type = SND_SEQ_EVENT_CONTROLLER;
	ev.data.control.channel = iChannel;
	ev.data.control.param   = iController;
	ev.data.control.value   = iValue;
	snd_seq_event_output_direct(m_pAlsaSeq, &ev);
}


// Receive MIDI controller event.
void qtractorMidiControlTestForm::receiveController ( unsigned short iChannel,
	unsigned short iController, unsigned short iValue )
{
	qDebug("receiveController: channel %u, controller %u, value %u",
		iChannel, iController, iValue);

	qtractorMidiControlObserver::dispatch(iChannel, iController, iValue);
}


// Local widget slots
void qtractorMidiControlTestForm::booleanCheckBoxChanged ( bool bValue )
{
	if (m_ui.BooleanCheckBox->observer()->isBusy())
		return;

	float fValue = (bValue ? 1.0f : 0.0f);
	qDebug("booleanCheckBoxChanged(%g)", fValue);
	m_ui.BooleanCheckBox->observer()->setValue(fValue);
}


void qtractorMidiControlTestForm::integerSpinBoxChanged ( const QString& sValue )
{
	if (m_ui.IntegerSpinBox->observer()->isBusy())
		return;

	float fValue = (sValue.toInt() / 100.0f);
	qDebug("integerSpinBoxChanged(%g)", fValue);
	m_ui.IntegerSpinBox->observer()->setValue(fValue);
}


void qtractorMidiControlTestForm::linearSpinBoxChanged ( const QString& sValue )
{
	if (m_ui.LinearSpinBox->observer()->isBusy())
		return;

	float fValue = sValue.toFloat();
	qDebug("linearSpinBoxChanged(%g)", fValue);
	m_ui.LinearSpinBox->observer()->setValue(fValue);
}


void qtractorMidiControlTestForm::linearSliderChanged ( int iValue )
{
	if (m_ui.LinearSlider->observer()->isBusy())
		return;

	float fValue = (float(iValue) / 10000.0f);
	qDebug("linearSliderChanged(%g)", fValue);
	m_ui.LinearSlider->observer()->setValue(fValue);
}


void qtractorMidiControlTestForm::logSpinBoxChanged ( const QString& sValue )
{
	if (m_ui.LogSpinBox->observer()->isBusy())
		return;

	float fValue = sValue.toFloat();
	qDebug("logSpinBoxChanged(%g)", fValue);
	m_ui.LogSpinBox->observer()->setValue(fValue);
}


void qtractorMidiControlTestForm::logSliderChanged ( int iValue )
{
	if (m_ui.LogSlider->observer()->isBusy())
		return;

	float fValue = (float(iValue) / 10000.0f);
	qDebug("logSliderChanged(%g)", fValue);
	m_ui.LogSlider->observer()->setValue(fValue);
}


// Timer slot funtion.
void qtractorMidiControlTestForm::timerSlot (void)
{
	// Always update observers ...
	qtractorSubject::flushQueue();

	// Register the next timer slot.
	QTimer::singleShot(QTRACTOR_TIMER_MSECS, this, SLOT(timerSlot()));
}


// end of qtractorMidiControlTestForm.cpp
